/*
GmailSend Library for MetaTrader
Copyright (C) 2008 OpenThinking Systems, LLC
http://www.openthinkingsystems.com
All rights reserved.

- Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

- Neither the name of OpenThinking Systems LLC nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/


#include "MailError.h"
#include "MailSender.h"
#include "inet_conn.h"
#include "ssl_cli.h"
#include "CyoEncode.h"

#include "convert.h"

#include <iostream>
#include <string>

#include <sys/types.h>
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif
#if HAVE_WINSOCK2_H
#include <winsock2.h>
#endif
#ifdef _WIN32
#include <windows.h>
#define inet_pton(x,y,z)
#else
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#endif


#define BUF_SIZE     1024

//constructors
MailSender::MailSender()
    : /*cntx(NULL),*/ smtp_server(""), smtp_port(0)
{
};

MailSender::MailSender(string const server, int const port)
    : smtp_server(server), smtp_port(port)
{
};

//API

/**
    Logs in to the user account on server using the username and password.

    \param[in]  u   Username
    \param[in]  p   Password

    \return     MSSuccess   On success
    \return     MSFailure   On error
*/
MailSenderError MailSender::Login(string u, string p)
{
    this->uname = u;
    this->pwd   = p;

    return MSSuccess;
}


/**
  Sends the 'email'.
 */
MailSenderError MailSender::Send(EMail &email)
{
    std::string serv_reply;
    std::string sender;
    std::string message;
    std::string recepient_TO;

    int         socket;
    int         tmp;
    char        cmd_to_host[BUF_SIZE];
    char        rsp_from_host[BUF_SIZE];

    SSL_CTX *ctx;
    SSL     *ssl;
    BIO     *sbio;

    // read the 'email' object and verify that it has
    // enough information for it to be sent

    // get email sender's email address
    if(email.GetSender(sender) == -1)
    {
        //cout<< "email.GetSender() failed\n";
        return MSFailure;
    }

    // get email content
    if(email.GetMessage(message) == -1)
    {
        //cout<< "email.GetMessage() failed\n";
        return MSFailure;
    }

    // atleast one recepient should be present
    if(email.GetFirstRecepientTo(recepient_TO) == -1)
    {
        //cout<< "email.GetFirstRecepientTo() failed\n";
        return MSFailure;
    }

    //cout << "opening.... " << (this->smtp_server).c_str() << "\n";
    // open socket connection with server
    socket = open_connection((this->smtp_server).c_str(), (int const)this->smtp_port);
    if(socket == -1)
    {
        //cout<< "open_connection() failed\n";
        return MSFailure;
    }

    // clear command and response buffers
    memset(cmd_to_host, 0, BUF_SIZE);
    memset(rsp_from_host, 0, BUF_SIZE);

    //while(1)
    {
        // wait for host's response
        tmp = recv(socket, rsp_from_host, BUF_SIZE, 0);
        if(tmp == -1)
        {
            //cout<< "recv returned -1, host closed connection\n";
            return MSFailure;
        }
        //else
        //{
            //fprintf(stdout, "host: ");
            //fprintf(stdout, "[received bytes: %d]\n", tmp);
            //cout<< std::string(rsp_from_host);
        //}
        memset(rsp_from_host, 0, BUF_SIZE);


        // prompt for next command
        //fprintf(stdout, "\nGO> ");

        // send HELO command to host
        tmp = send(socket, "HELO\n", strlen("HELO\n"), 0);
        if(tmp == -1)
        {
            //cout<< "send() returned -1\n";
            return MSFailure;
        }

        // wait for host's response
        tmp = recv(socket, rsp_from_host, BUF_SIZE, 0);
        if(tmp == -1)
        {
            //cout<< "recv returned -1, host closed connection\n";
            return MSFailure;
        }
        //else
        //{
            //fprintf(stdout, "host: ");
            //fprintf(stdout, "[received bytes: %d]\n", tmp);
        //    cout<< std::string(rsp_from_host);
       // }
        memset(rsp_from_host, 0, BUF_SIZE);

        // send STARTTLS command
        tmp = send(socket, "STARTTLS\n", strlen("STARTTLS\n"), 0);
        if(tmp == -1)
        {
            //cout<< "send() returned -1\n";
            return MSFailure;
        }

        // wait for host's response
        tmp = recv(socket, rsp_from_host, BUF_SIZE, 0);
        if(tmp == -1)
        {
            //cout<< "recv returned -1, host closed connection\n";
            return MSFailure;
        }
        //else
        //{
            //fprintf(stdout, "host: ");
            //fprintf(stdout, "[received bytes: %d]\n", tmp);
        //    cout<< std::string(rsp_from_host);
        //}
        memset(rsp_from_host, 0, BUF_SIZE);

	//cout<< "***Starting TLS handshake***\n";
        // Build our SSL context
        ctx = initialize_ctx(KEYFILE, PASSWORD);

        // TLS session handshake

        // Connect the SSL socket
        ssl  = SSL_new(ctx);
        sbio = BIO_new_socket(socket, BIO_NOCLOSE);
        SSL_set_bio(ssl, sbio, sbio);

        if(SSL_connect(ssl) <= 0)
            berr_exit("SSL connect error");
        /*
        if(require_server_auth)
            check_cert(ssl,host);
        */

	//cout<< "***SSL handshake done***\n";

        // send mail through TLS
        this->send_secure_mail(email, ssl);

        // close TLS session
	SSL_shutdown(ssl);
        SSL_free(ssl);
        destroy_ctx(ctx);
        //return(0);
	//cout<< "***SSL session terminated***\n";

        // close internet socket connection
        close(socket);
	//cout<< "***socket closed***\n";
    }

    return MSSuccess;
}

// local non-class "helper" function to send
// data through ssl and check sent status for error codes
// Returns 0 on success, exits program on error :-O
static int ssl_send(string s, SSL *ssl)
{
    int r;

    // send data
    r = SSL_write(ssl, s.c_str(), strlen(s.c_str()));
    switch(SSL_get_error(ssl, r))
    {
        case SSL_ERROR_NONE:
            if(strlen(s.c_str()) != r)
                err_exit("Incomplete write!");
        break;
        default:
            berr_exit("SSL write problem");
    }

    return 0;
}

// local non-class "helper" function to receive
// data through ssl and check receive status for error codes
// Returns >0 on success (number of bytes read), exits program on error :-O
// On success, "s" is loaded with the received message
static int ssl_recv(char *buf, int buf_len, SSL *ssl)
{
    int r;

    // Read the server's response
    //while(1)
    {
        r = SSL_read(ssl, buf, buf_len);

        if(r > 0)
        {
            //cout<< "Read " << r << " bytes\n";
            return r;
        }

        switch(SSL_get_error(ssl, r))
        {
            case SSL_ERROR_NONE:
                //cout<< "SSL_ERROR_NONE: but SSL_read read 0 or less bytes!\n";
                return 0;
            break;

            /*
            case SSL_ERROR_ZERO_RETURN:
            {
                r = SSL_shutdown(ssl);
                switch(r)
                {
                    case 1:
                    break; //Success
                    case 0:
                    case -1:
                    default:
                        berr_exit("Shutdown failed");
                    break;
                }
            }
            break;
            */

            case SSL_ERROR_SYSCALL:
                //cout<< "SSL Error: Premature close\n";
                SSL_free(ssl);
                //free(request);
                exit(0);

            default:
                berr_exit("SSL read problem");
            break;
        }

        return 0;
    }
}

/**
    Send email through the already established SSL connection.
    \param[in] email    The email object.
    \param[in] ssl      The ssl connection object.

    \return    0        On success
    \return    -1       On failure

    Sends email by issuing the commands MAIL, RCPT and DATA.
*/
int     MailSender::send_secure_mail(EMail& email, SSL *ssl)
{
#define RSP_BUF_SIZE 1000
    string  request, tmp;
    char    response[RSP_BUF_SIZE];
    char    temp_buf[100];
    int     rsp_len, response_numeric;
    char    crlf[3] = "\r\n"; //{13, 10, 0};

    // send HELO command
    request = string("HELO") + string(crlf);
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

    // perform SMTP AUTH using the LOGIN method
    // send AUTH command
    request = string("AUTH LOGIN") + string(crlf);
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

    // send 'username' in base64
    memset(temp_buf, 0, sizeof(temp_buf));
    //temp_buf = new (sizeof(char) * CyoEncode::Base64EncodeGetLength(strlen(this->uname)));
    CyoEncode::Base64Encode(temp_buf, (this->uname).c_str(), strlen((this->uname).c_str()));
    request = string(temp_buf);
    request = request + crlf;
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }


    // send 'password' in base64
    memset(temp_buf, 0, sizeof(temp_buf));
    //temp_buf = new (sizeof(char) * CyoEncode::Base64EncodeGetLength(strlen(this->pwd)));
    CyoEncode::Base64Encode(temp_buf, (this->pwd).c_str(), strlen((this->pwd).c_str()));
    request = string(temp_buf);
    request = request + crlf;
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
   //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

/*
    // send RSET command to ensure past garbage is discarded
    request = "RSET";
    request = request + crlf;
    cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        cout<< "Error response \n";
        // error
        return -1;
    }
*/

    // send MAIL command
    request = "MAIL FROM:<";
    email.GetSender(tmp);
    request = request + tmp + ">" + crlf;
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

#if 1
    //while loop for all "TO recepients"

    email.GetFirstRecepientTo(tmp);
        //cout<< "***GetFirstRecepientTo() failed***\n";

    while(tmp != "")
    {
        // send RCPT command(s)
        request = "RCPT TO:<";
        request = request + tmp + ">" + crlf;
        //cout<< "\nSending command: " << request;
        ssl_send(request, ssl);

        // read response
        memset(response, 0, RSP_BUF_SIZE);
        rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
        //cout<< "Received: " << response << "\n";

        // convert response to numeric form
        response[3] = '\0';
        response_numeric = atoi(response);
        //cout<< "response_numeric: " << response_numeric << "\n";
        if( response_numeric >= 421 )
        {
            //cout<< "Error response \n";
            // error
            return -1;
        }

        if(email.GetNextRecepientTo(tmp) == -1)
        {
            //cout<< "***GetNextRecepientTo() failed***\n";
        }

    } // while loop for all "TO recepients"

    // send DATA command
    request = string("DATA") + string(crlf);
    //cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

    // send "email content"
    email.GetMessage(tmp);
        //cout<< "***GetMessage() failed***\n";

    request = tmp;
    request = request + crlf + "." + crlf;
    //cout<< "\nSending command (email content):\n" << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    //cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    //cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        //cout<< "Error response \n";
        // error
        return -1;
    }

/*
    // send SOML command
    request = "SOML\n";
    cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        cout<< "Error response \n";
        // error
        return -1;
    }
*/

/*
    // send QUIT command
    request = string("QUIT") + string(crlf);
    cout<< "\nSending command: " << request;
    ssl_send(request, ssl);

    // read response
    memset(response, 0, RSP_BUF_SIZE);
    rsp_len = ssl_recv(response, RSP_BUF_SIZE-1, ssl);
    cout<< "Received: " << response << "\n";

    // convert response to numeric form
    response[3] = '\0';
    response_numeric = atoi(response);
    cout<< "response_numeric: " << response_numeric << "\n";
    if( response_numeric >= 421 )
    {
        cout<< "Error response \n";
        // error
        return -1;
    }
*/
#endif // 0/1

    return 0;
}
